package main

import (
	"fmt"

	"github.com/urfave/cli/v2"
	"github.com/urfave/cli/v2/altsrc"
)

var (
	configFileFlag = &cli.StringFlag{
		Name:    "config",
		Aliases: []string{"c"},
		Usage:   "The path to the config file",
		Value:   "config.toml",
	}

	dataDirFlag = &cli.StringFlag{
		Name:    "data-dir",
		Aliases: []string{"d"},
		Usage:   "The path to the data directory",
		Value:   "data",
	}

	metricsListenAddrFlag = &cli.StringFlag{
		Name:  "metrics-listen",
		Usage: "The listen address of the metrics server",
		Value: "0.0.0.0:9429",
	}

	logLevelFlag = &cli.StringFlag{
		Name:  "log-level",
		Usage: "The log level",
		Value: "info",
	}

	grpcListenAddrFlag = &cli.StringFlag{
		Name:  "grpc-listen",
		Usage: "The listen address of the grpc server",
		Value: "0.0.0.0:9430",
	}

	rpcListenAddrFlag = &cli.StringFlag{
		Name:  "rpc-listen",
		Usage: "The listen address of the jsonrpc server",
		Value: "0.0.0.0:9431",
	}

	privateKeyFlag = &cli.StringFlag{
		Name:    "private-key",
		Usage:   "The private key of the account",
		EnvVars: []string{"PRIVATE_KEY"},
	}

	chainRPCFlag = &cli.StringFlag{
		Name:  "chain-rpc",
		Usage: "The host of the chain json rpc server",
	}

	storeContractFlag = &cli.StringFlag{
		Name:  "store-contract",
		Usage: "The address of the store contract",
	}

	witnessContractFlag = &cli.StringFlag{
		Name:  "witness-contract",
		Usage: "The address of the reward contract",
	}

	commitTimeFlag = &cli.IntFlag{
		Name:  "commit-time",
		Usage: "The time to commit the proof",
	}

	commitExpireFlag = &cli.IntFlag{
		Name:  "commit-expire",
		Usage: "The time to expire the proof",
	}
)

// p2p flags
var (
	p2pUseDiscoveryFlag = &cli.BoolFlag{
		Name:  "p2p-use-discovery",
		Usage: "Whether to use discovery",
		Value: true,
	}

	p2pStaticPeersFlag = &cli.StringSliceFlag{
		Name:  "p2p-static-peers",
		Usage: "Static peers",
	}

	p2pBootstrapNodeAddrFlag = &cli.StringSliceFlag{
		Name:  "p2p-bootstrap-node-addr",
		Usage: "Bootstrap node address",
	}

	p2pLocalIPFlag = &cli.StringFlag{
		Name:  "p2p-local-ip",
		Usage: "The local ip address to listen for incoming data.",
		Value: "",
	}

	p2pHostFlag = &cli.StringFlag{
		Name:  "p2p-host-ip",
		Usage: "The IP address advertised by libp2p. This may be used to advertise an external IP.",
		Value: "",
	}

	p2pPrivKeyFlag = &cli.StringFlag{
		Name:    "p2p-private-key",
		Usage:   "The file path of the private key to use in communications with other peers.",
		EnvVars: []string{"P2P_PRIVATE_KEY"},
	}

	p2pTCPPortFlag = &cli.IntFlag{
		Name:  "p2p-tcp-port",
		Usage: "The port used by libp2p.",
		Value: 30333,
	}

	p2pUDPPortFlag = &cli.IntFlag{
		Name:  "p2p-udp-port",
		Usage: "The port used by discv5.",
		Value: 30334,
	}

	p2pMaxPeersFlag = &cli.IntFlag{
		Name:  "p2p-max-peers",
		Usage: "The max number of p2p peers to maintain.",
		Value: 16,
	}

	p2pMaxInboundPeersFlag = &cli.IntFlag{
		Name:  "p2p-max-inbound-peers",
		Usage: "The max number of inbound p2p peers to maintain.",
		Value: 10,
	}

	p2pMaxOutboundPeersFlag = &cli.IntFlag{
		Name:  "p2p-max-outbound-peers",
		Usage: "The max number of outbound p2p peers to maintain.",
		Value: 10,
	}

	ignoreLocalIPFlag = &cli.BoolFlag{
		Name:  "ignore-local-ip",
		Usage: "The node is not connected to the local IP address",
		Value: false,
	}
)

func loadFlagsFromConfig(cliCtx *cli.Context, flags []cli.Flag) error {
	if cliCtx.IsSet(configFileFlag.Name) {
		return altsrc.InitInputSourceWithContext(
			flags,
			altsrc.NewTomlSourceFromFlagFunc(configFileFlag.Name),
		)(cliCtx)
	}
	return nil
}

// WrapFlags so that they can be loaded from alternative sources.
func wrapFlags(flags []cli.Flag) []cli.Flag {
	wrapped := make([]cli.Flag, 0, len(flags))
	for _, f := range flags {
		switch t := f.(type) {
		case *cli.BoolFlag:
			f = altsrc.NewBoolFlag(t)
		case *cli.DurationFlag:
			f = altsrc.NewDurationFlag(t)
		case *cli.GenericFlag:
			f = altsrc.NewGenericFlag(t)
		case *cli.Float64Flag:
			f = altsrc.NewFloat64Flag(t)
		case *cli.IntFlag:
			f = altsrc.NewIntFlag(t)
		case *cli.StringFlag:
			f = altsrc.NewStringFlag(t)
		case *cli.StringSliceFlag:
			f = altsrc.NewStringSliceFlag(t)
		case *cli.Uint64Flag:
			f = altsrc.NewUint64Flag(t)
		case *cli.UintFlag:
			f = altsrc.NewUintFlag(t)
		case *cli.PathFlag:
			f = altsrc.NewPathFlag(t)
		case *cli.Int64Flag:
			// Int64Flag does not work. See https://github.com/prysmaticlabs/prysm/issues/6478
			panic(fmt.Sprintf("unsupported flag type type %T", f))
		case *cli.IntSliceFlag:
			f = altsrc.NewIntSliceFlag(t)
		default:
			panic(fmt.Sprintf("cannot convert type %T", f))
		}
		wrapped = append(wrapped, f)
	}
	return wrapped
}
