package tree

import (
	"bytes"
	"math/big"

	"github.com/ethereum/go-ethereum/crypto"
)

var (
	LeafPrefix = []byte{0x00}
	NodePrefix = []byte{0x01}
)

type MerkleSumTree struct {
	sideNodes [][32]byte
	nodeSums  []*big.Int
}

func leafDigest(data [32]byte, val *big.Int) (digest [32]byte) {
	valBytes := make([]byte, 32)
	copy(valBytes[32-len(val.Bytes()):], val.Bytes())
	buf := bytes.NewBuffer(LeafPrefix)
	buf.Write(valBytes)
	buf.Write(data[:])
	copy(digest[:], crypto.Keccak256Hash(buf.Bytes()).Bytes())
	return
}

func nodeDigest(leftDigest, rightDigest [32]byte, leftVal, rightVal *big.Int) (digest [32]byte) {
	leftValBytes := make([]byte, 32)
	rightValBytes := make([]byte, 32)

	copy(leftValBytes[32-len(leftVal.Bytes()):], leftVal.Bytes())
	copy(rightValBytes[32-len(rightVal.Bytes()):], rightVal.Bytes())

	buf := bytes.NewBuffer(NodePrefix)
	buf.Write(leftValBytes)
	buf.Write(leftDigest[:])
	buf.Write(rightValBytes)
	buf.Write(rightDigest[:])
	copy(digest[:], crypto.Keccak256Hash(buf.Bytes()).Bytes())
	return
}

func ComputeMST(dataList [][32]byte, values []*big.Int) (nodes [][32]byte, sums []*big.Int) {
	nodeDigests := make([][32]byte, len(dataList))
	nodeSums := make([]*big.Int, len(dataList))

	for i := 0; i < len(dataList); i++ {
		nodeDigests[i] = leafDigest(dataList[i], values[i])
		nodeSums[i] = values[i]
	}

	odd := len(dataList) & 1
	size := (len(dataList) + 1) >> 1

	pNodeDigests := make([][32]byte, len(dataList))
	pNodeSums := make([]*big.Int, len(dataList))

	copy(pNodeDigests, nodeDigests)
	copy(pNodeSums, nodeSums)

	for {
		i := 0
		for i < size-odd {
			j := i << 1
			nodeDigests[i] = nodeDigest(pNodeDigests[j], pNodeDigests[j+1], pNodeSums[j], pNodeSums[j+1])
			nodeSums[i] = new(big.Int).Add(pNodeSums[j], pNodeSums[j+1])
			i++
		}

		if odd == 1 {
			nodeDigests[i] = pNodeDigests[i<<1]
			nodeSums[i] = pNodeSums[i<<1]
		}

		if size == 1 {
			break
		}

		odd = size & 1
		size = (size + 1) >> 1
		copy(pNodeDigests, nodeDigests)
		copy(pNodeSums, nodeSums)
	}
	return nodeDigests, nodeSums
}
